<?php
class DcApi
{
    protected $gitRepo = '';
    protected $username = '';
    protected $gitServer = '';
    protected $gitBranch = '';
    protected $gitAuthor = '';
    protected $gitEmail = '';
    protected $dcSouces = array();
    protected $userEmail = ''; // for git  commit logs
    protected $dcSettings = null;

    function __construct($data = array())
    {
        $this->dcSettings = new DcSettings();

        if (!empty($data))
        {

            $this->initialize($data);
        }
    }

    function initialize($config = array())
    {

        syslog(LOG_DEBUG, 'Initializing DCAPI...');


        foreach ($config as $key => $val)
        {
            if (isset($this->$key))
            {
                $method = 'set_' . $key;

                if (method_exists($this, $method))
                {
                    $this->$method($val);
                }
                else
                {
                    $this->$key = $val;
                    syslog(LOG_DEBUG, sprintf("setting %s => %s in git service class", $key, $val));
                }
            }
        }
    }

    function getUsername()
    {

        return $this->username;
    }

    function getGitAuthorName()
    {

        return $this->gitAuthor;
    }

    function getGitAuthorEmail()
    {

        return $this->gitEmail;
    }

    function getGitServer()
    {

        return $this->gitServer;
    }

    function getGitBranch()
    {
        if (empty($this->gitBranch))
        {
            $this->gitBranch = 'master';
        }

        return $this->gitBranch;
    }

    function getGitRepo()
    {

        return $this->gitRepo;
    }

    function getUserEmail()
    {

        return $this->userEmail;
    }


    function getGitHash()
    {
        syslog(LOG_DEBUG,"Retrieving git revision hash");
        $username = $this->getUsername();
        $repoPath = $this->dcSettings->get_dc_staging_repository();
        syslog(LOG_DEBUG, "Checking for staging repository at ::  $repoPath");

        try
        {
            $repo = new CF_GitRepo($repoPath, false, false, $username);
            return $repo->get_git_revision();
        }
        catch(Exception $e)
        {

            throw new DC_Exception("Error getting GIT revision of the staging repository at ".$repoPath);
        }


    }

    function checkForGitKeys()
    {
        $keyPath = $this->dcSettings->get_dc_user_ssh_key_directory($this->getUsername());
        $keyname = $this->dcSettings->get_dc_user_ssh_key_name();
        $file = $keyPath . DIRECTORY_SEPARATOR . $keyname;

        return file_exists($file);
    }

    function gitInitStagingFromRemote($scriptData)
    {
        syslog(LOG_DEBUG, 'Initializing git clone for staging area..');
        $username = $this->getUsername();
        $gitServerUrl = $scriptData['gitServer'];
        $gitBranch = empty($scriptData['gitBranch']) ? 'master' : $scriptData['gitBranch'];
        $cloneDir = $this->dcSettings->get_dc_staging_repository();
        $this->_makeStageRepositoryBackup($cloneDir);

        try
        {
            $repo = new CF_GitRepo($cloneDir, true, false, $username);
            $repo->clone_remote_mirror($gitServerUrl, $gitBranch);
            $this->_generateUpdateScripts($scriptData);
        }
        catch(Exception $e)
        {

            throw new Exception('Error setting up new  clone repository for staging area. ' . $e->getMessage());
        }
    }

    function _makeStageRepositoryBackup($cloneDir)
    {

        $backupDir = $this->dcSettings->get_dc_stage_repo_backup_path();
        $backupDirName = $backupDir . DIRECTORY_SEPARATOR . date('Y-m-d-H-i');
        syslog(LOG_DEBUG, 'Backing up staging area to new location ::' . $backupDirName);
        if (!@file_exists($backupDir))
        {
            $oldumask = umask(0);
            $result = @mkdir($backupDir, '0755');
            umask($oldumask);
            if (!$result)
            {
                syslog(LOG_DEBUG, sprintf('Error creating backup directory to %s ::', $backupDirName));

                throw new Exception('Error creating backup directory for staging repository in location ' . $backupDirName);
            }
        }
        if (@file_exists($cloneDir))
        {
            $renameResult = rename($cloneDir, $backupDirName);
            if (!$renameResult)
            {
                syslog(LOG_ERR, sprintf('Backing up Failed from directory  %s to %s ::', $cloneDir, $backupDirName));
            }
        }
    }

    function gitInitStagingFromPath($path, $username)
    {
        syslog(LOG_DEBUG, "Initializing git clone for staging area from given path: $path");
        $cloneDir = $this->dcSettings->get_dc_staging_repository();

        try
        {
            if (empty($path))
            {

                throw new Error("Path cannot be empty");
            }
            if (is_dir($path) && !is_dir_empty($path))
            {
                @copyr($path, $cloneDir);
                // initialize a git repo here
                Utills::deleteAll($cloneDir . DIRECTORY_SEPARATOR . '.git', false);
                $repo = new CF_GitRepo($cloneDir, true, true, $username);
                $repo->add();
                $repo->commit("Initializing staging repository", 'admin', 'admin@cfengine.com', 'admin@cfengine.com');
                $repo->create_branch($this->dcSettings->get_dc_git_working_branch());
                delete_git_update_scripts();
            }
            else
            {

                throw new Exception("Path supplied is either not a directory or is empty.");
            }
        }
        catch(Exception $e)
        {

            throw new Exception('Error setting up new  clone repository for staging area.' . $e->getMessage());
        }
    }

    function _generateUpdateScripts($data)
    {
        $this->_generateScriptFromTemplates('params.template', 'params.sh', $data);
    }

    function _generateScriptFromTemplates($templateName, $filename, $data)
    {

        $scriptTemplatePath = $this->dcSettings->get_dc_update_script_template_path() . DIRECTORY_SEPARATOR . $templateName;
        $templateContent = @file_get_contents($scriptTemplatePath);
        if ($templateContent === false)
        {
            syslog(LOG_ERR, "Template for $templateName script not found at path:" . $scriptTemplatePath);

            throw new Exception("Cannot get the content of the $templateName script template to generate the script from path $scriptTemplatePath");
        }
        else
        {
            $replace = $this->_getScriptTemplateParameters($data);
            $generatedScript = str_replace(array_keys($replace) , array_values($replace) , $templateContent);
            $scriptPath = $this->dcSettings->get_dc_update_script_path() . DIRECTORY_SEPARATOR . $filename;
            syslog(LOG_DEBUG, "writing $templateName script at path: $scriptPath");

            $result = @file_put_contents($scriptPath, $generatedScript);
            if ($result === false)
            {
                syslog(LOG_ERR, "cannot write $filename script at path: $scriptPath");

                throw new Exception("Cannot generate $filename script file at $scriptPath");
            }
            else
            {
                chmod($scriptPath, 0750);
            }
        }
    }

    function _getScriptTemplateParameters($data)
    {
        $gitServerUrl = $data['gitServer'];
        $gitBranch = $data['gitBranch'];
        $gitEmail = $data['gitEmail'];
        $gitAuthor = $data['gitAuthor'];

        $keyPath = $this->dcSettings->get_dc_user_ssh_key();
        $workDir = $this->dcSettings->get_dc_staging_repository();
        $gitLocalWorkingBranch = $this->dcSettings->get_dc_git_working_branch();
        $scriptDir = $this->dcSettings->get_dc_update_script_template_path();

        $replace = array(
            '%WORKING_DIR%' => $workDir,
            '%REMOTE_GIT_URL%' => $gitServerUrl,
            '%REMOTE_GIT_BRANCH%' => $gitBranch,
            '%LOCAL_GIT_BRANCH%' => $gitLocalWorkingBranch,
            '%SSH_PRIVATE_KEY%' => $keyPath,
            '%GIT_EMAIL%' => $gitEmail,
            '%GIT_AUTHOR%' => $gitAuthor,
            '%VCS_TYPE%' => 'GIT',
            '%SCRIPT_DIR%' => $scriptDir
        );

        foreach ($replace as $key => $val)
        {
            $replace[$key] = '"' . str_replace('"','\"',$val) . '"';
        }

        return $replace;
    }

    function getExistingGitRepo()
    {

        $username = $this->getUsername();
        $repoPath = $this->dcSettings->get_dc_user_working_directory($username);
        $repo = new CF_GitRepo($repoPath, false, false, $username);
        syslog(LOG_DEBUG, "Getting git repo on path $repoPath");

        return $repo;
    }

    function checkStagingRepository()
    {

        $username = $this->getUsername();
        $repoPath = $this->dcSettings->get_dc_staging_repository();
        syslog(LOG_DEBUG, "Checking for staging repository at ::  $repoPath");

        try
        {
            $repo = new CF_GitRepo($repoPath, false, false, $username);
        }
        catch(Exception $e)
        {

            throw new DC_Staging_Notfound_Exception("Staging repository not found");
        }

        return $repo;
    }

    function checkUserRepository()
    {

        $username = $this->getUsername();
        $gitServerUrl = $this->getGitServer();
        $workDir = $this->dcSettings->get_dc_user_working_directory($this->getUsername());
        $gitBranch = $this->getGitBranch();
        $keyPath = $this->dcSettings->get_dc_user_ssh_key_directory($this->getUsername());

        syslog(LOG_DEBUG, "Checking for user repository at :: $workDir");

        try
        {
            $repo = new CF_GitRepo($workDir, false, false, $username);
            $repo->reset_hard($this->dcSettings->get_dc_git_working_branch());
        }
        catch(Exception $e)
        {
            // no repository found.Initialize one

            try
            {
                if (!file_exists($workDir))
                {

                    throw new Exception("No working directory found on path $workDir", 1);
                }
                Utils::deleteAll($workDir);
                $repo = new CF_GitRepo($workDir, true, false, $username);
                $repo->clone_remote($gitServerUrl);
                $repo->checkout_remote_branch($this->dcSettings->get_dc_git_working_branch());
            }
            catch(Exception $e)
            {
                $message = sprintf("Failed cloning repository with the url:: %s on path %s.
                                        %s", $gitServerUrl, $workDir, $e->getMessage());
                syslog(LOG_ERR, 'Failed cloning repository :: ' . $e->getMessage());

                throw new Exception($message);
            }
        }
    }

    function initializeWorkSpace($username)
    {

        $workDir = $this->dcSettings->get_dc_user_working_directory($username);
        $keyPath = $this->dcSettings->get_dc_user_ssh_key_directory($username);
        if (!is_dir($workDir))
        {
            $result = @mkdir($workDir, 0700, true);
            if (!$result)
            {

                throw new Exception("cannot create workdir on path: $workDir");
            }
        }

        if (!is_dir($keyPath))
        {
            $keyResult = @mkdir($keyPath, 0700, true);
            if (!$keyResult)
            {

                throw new Exception("cannot create key store on path:  $keyPath");
            }
        }
        $this->generateApiConfigFile($username);
    }

    function generateApiConfigFile($username)
    {
        /**
         *  {
         * - log: "STDERR",
         * - log_level: 4,
         * - repolist: [ "../sketches" ],
         * - recognized_sources: [ "/var/cfengine/design-center/sketches" ],
         * - runfile: { location: "api-runfile.cf", standalone: false, relocate_path: "sketches" },
         * - vardata: "vardata.conf",
         * -}
         *
         */
        $repoListLocation = $this->dcSettings->get_dc_user_repolist_location($username);
        $recognizedSources = $this->dcSettings->get_dc_sketch_recognized_sources();
        $runfileLocation = $this->dcSettings->get_dc_sketch_runfile_location($username);
        $vardataLocation = $this->dcSettings->get_dc_sketch_vardatafile_location($username);
        $apiFilePath = $this->dcSettings->get_dc_user_sketch_config_file_path($username);
        $constdata = $this->dcSettings->get_dc_sketch_const_data();

        $configStructure = array(
            "log" => "STDERR",
            "log_level" => "3",
            "repolist" => array(
                $repoListLocation
            ) ,
            "recognized_sources" => $recognizedSources,
            "constdata" => $constdata,
            "runfile" => array(
                "location" => $runfileLocation
            ) ,
            "vardata" => $vardataLocation
        );
        $json = json_encode($configStructure);
        $result = file_put_contents($apiFilePath, $json);
    }

    function resetGitRepository()
    {
        $repo = $this->getExistingGitRepo();
        $repo->reset_hard($this->getGitBranch());
    }

    function commit_push($msg, $authorName, $authorEmail, $userEmail)
    {

        $repo = $this->getExistingGitRepo();
        $repo->add();
        $repo->commit($msg, $authorName, $authorEmail, $userEmail);
        $gitUrl = $this->getGitServer();
        $branch = $this->getGitBranch();
        $repo->push('origin', $branch);
        $repo->post_push();
    }

    function reset_and_pull()
    {
        $repo = $this->getExistingGitRepo();
        $repo->pre_fetch();
        $repo->fetch();
        $repo->reset_hard($this->getGitBranch());
    }
}

class DC_Exception extends Exception
{
}
class DC_Settings_Exception extends DC_Exception
{
}; // when settings are not specified
class DC_Staging_Notfound_Exception extends DC_Exception
{
};
class Git_Settings_Exception extends DC_Exception
{
}; // related to git errors


?>