<?php

/**
 * Class InventoryLib
 */
class InventoryLib
{
    /**
     * @var VariablesDictionaryModel
     */
    private $variableDictionaryModel;


    public function __construct(VariablesDictionaryModel $variablesDictionary)
    {
        $this->variableDictionaryModel = $variablesDictionary;
    }

    /**
     * @var int
     */
    private $limit = 1000;
    /**
     * @var int
     */
    private $maxLimit = 1000000000;

    /**
     * @var array
     */
    private $operatorsMap = [
        //Numbers
        '<' => ['operator' => '<', 'function' => 'cf_convertToNum'],
        '>' => ['operator' => '>', 'function' => 'cf_convertToNum'],
        '=' => ['operator' => '=', 'function' => 'cf_convertToNum'],
        '!=' => ['operator' => '!=', 'function' => 'cf_convertToNum'],
        '<=' => ['operator' => '<=', 'function' => 'cf_convertToNum'],
        '>=' => ['operator' => '>=', 'function' => 'cf_convertToNum'],
        //Strings
        'matches' => ['operator' => 'ILIKE'],
        'not_match' => ['operator' => 'NOT ILIKE'],
        'contains' => ['operator' => 'LIKE'],
        'not_contain' => ['operator' => 'NOT LIKE'],
    ];


    /**
     * @param $data
     * @return string
     * @throws Exception
     */
    public function applyFilter($data)
    {

        $where = ['WHERE'];
        if (isset($data->filter)) {
            $dataFilter = $data->filter;
            foreach ($dataFilter as $variable => $opVal) {
                foreach ($opVal as $operator => $value) {
                    if (is_array($value)) {
                        foreach ($value as $v) {
                            $where[] = sizeof($where) > 1 ? 'AND' : '';
                            $where[] = $this->makeWhereCondition($operator, $variable, $v);
                        }
                    } else {
                        $where[] = sizeof($where) > 1 ? 'AND' : '';
                        $where[] = $this->makeWhereCondition($operator, $variable, $value);
                    }
                }
            }
        }

        //if $where contains only 1 element it's WHERE, skip it
        return sizeof($where) > 1 ? implode(' ', $where) : '';
    }


    /**
     * @param $data
     * @return string
     */
    public function applySorting($data)
    {
        if (empty($data->sort)) {
            return '';
        }
        $sort = $data->sort;
        if ($sort[0] == '-')  //check first symbol
        {
            $variable = substr($sort, 1); //if first symbol is `-`, get variable without `-`
            $direction = 'DESC';
        } else {
            $variable = $sort;
            $direction = 'ASC';
        }

        $variableEntity = $this->variableDictionaryModel->getByName($variable);

        if ($variableEntity && in_array($variableEntity['type'], ['int', 'real'])) {
            return " ORDER BY cf_convertToNum(values->>'$variable') $direction ";
        }

        return " ORDER BY values->>'$variable' $direction ";

    }


    /**
     * @param $data
     * @return string
     * @throws Exception
     */
    public function applySelect($data)
    {
        $variableModel = new VariablesDictionaryModel();
        if (!isset($data->select)) {
            throw new Exception('Select field is required');
        }

        if (!is_array($data->select)) {
            throw new Exception('Wrong Select data');
        }

        $selectArray = $data->select;

        $selectData = [];
        $functionsMap = $variableModel->getFunctionsMap();

        foreach ($selectArray as $key => $select) {
            if($select == 'hostkey'){
                $selectData[] = " $select AS \"$select\"";
            }
            elseif ($select == 'resultCount'){
                $selectData[] = " count(*) as count ";
            }
            elseif ($functionsMap[$select]) {
                $selectData[] = " {$functionsMap[$select]}(values->>'$select') AS \"$select\"";
            } else {
                $selectData[] = " values->>'$select' AS \"$select\"";
            }


        }
        return implode(', ', $selectData);
    }


    /**
     * @param $literalOperator
     * @param $variable
     * @param $value
     * @return string
     * @throws Exception
     */
    private function makeWhereCondition($literalOperator, $variable, $value)
    {
        if (!isset($this->operatorsMap[$literalOperator])) {
            throw new Exception('Wrong abbreviation of operator');
        }

        if($variable == 'hostkey'){
            return vsprintf("%s %s '%s'", [$variable, 'LIKE', $value]);
        }

        $conditionData = $this->operatorsMap[$literalOperator];
        $variableEntity = $this->variableDictionaryModel->getByName($variable);
        $value = $this->prepareValueToWhere($value, $literalOperator, $variableEntity['type']);

        return
            isset($conditionData['function']) && in_array($variableEntity['type'], ['int', 'real']) ?

                vsprintf("%s(values->>'%s') %s %s(%s)",
                    [
                        $conditionData['function'],
                        $variable,
                        $conditionData['operator'],
                        $conditionData['function'],
                        $value
                    ]) :
                vsprintf("values->>'%s' %s %s", [$variable, $conditionData['operator'], $value]);
    }

    private function prepareValueToWhere($value, $literalOperator, $type)
    {
        $filterValueFormatter = FormatInventoryFilterFactory::getImplementation($type);
        return $filterValueFormatter->getValueForWhere($value, $literalOperator);
    }


    /**
     * @param $data
     * @param $showSQL
     * @param HttpClientBase $httpClient
     * @return Response
     */
    public function sendRequestToQueryApi($data, HttpClientBase $httpClient, $showSQL = false)
    {
        $requestHeaders = getallheaders();
        $headers = ['Authorization: ' . $requestHeaders['Authorization']];
        return $this->changeResponse($httpClient->post('/query', $data, $headers), $showSQL);

    }

    /**
     * @param $data
     * @return string
     */
    private function changeResponse($data, $showSQL = false)
    {
        $data = json_decode($data);
        if (isset(reset($data->data)->query) AND !$showSQL) {
            unset(reset($data->data)->query);
        }
        return json_encode($data, JSON_PRETTY_PRINT);
    }

    /**
     * @param $limit
     * @return int
     */
    public function getLimit($limit = 1000)
    {
        //if the limit is null then set -1 as non-limit value
        $limit = $limit === null ? -1 : $limit;
        $limit = (isset($limit) && is_numeric($limit)) ? $limit : $this->limit;
        return $limit > $this->maxLimit ? $this->maxLimit : $limit;
    }

}