<?php

if (!defined('BASEPATH'))
    exit('No direct script access allowed');

class git_service
{

    protected $gitRepo = '';
    protected $username = '';
    protected $gitServer = '';
    protected $gitBranch = '';
    protected $gitAuthor = '';
    protected $gitEmail = '';
    protected $dcSouces = array();
    protected $userEmail = ''; // for git  commit logs

    function __construct($data = array())
    {
        if (!empty($data))
        {
            $this->initialize($data);
        }
    }

    function initialize($config = array())
    {
        log_message('debug','Initializing git service ..');
        foreach ($config as $key => $val)
        {
            if (isset($this->$key))
            {
                $method = 'set_' . $key;

                if (method_exists($this, $method))
                {
                    $this->$method($val);
                }
                else
                {
                    $this->$key = $val;
                    log_message('debug',sprintf("setting %s => %s in git service class", $key ,$val));
                }
            }
        }
    }

    function getUsername()
    {
        return $this->username;
    }

    function getGitAuthorName()
    {
        return $this->gitAuthor;
    }

    function getGitAuthorEmail()
    {
        return $this->gitEmail;
    }

    function getGitServer()
    {
        return $this->gitServer;
    }

    function getGitBranch()
    {
        if (empty($this->gitBranch))
        {
            $this->gitBranch = 'master';
        }
        return $this->gitBranch;
    }

    function getGitRepo()
    {
        return $this->gitRepo;
    }

    function getUserEmail()
    {
        return $this->userEmail;
    }

    function checkForGitKeys()
    {
        $keyPath = get_dc_user_ssh_key_directory($this->getUsername());
        $keyname = get_dc_user_ssh_key_name();
        $file = $keyPath . DIRECTORY_SEPARATOR . $keyname;

        return file_exists($file);
    }

    function gitInitStagingFromRemote()
    {
        log_message('debug', 'Initializing git clone for staging area..');
        $username = $this->getUsername();
        $gitServerUrl = $this->getGitServer();
        $gitBranch = $this->getGitBranch();

        $cloneDir = get_dc_staging_repository();
        $this->_makeStageRepositoryBackup($cloneDir);
        try
        {
            $repo = new CF_GitRepo($cloneDir, true, false, $username);
            $repo->clone_remote_mirror($gitServerUrl,$gitBranch);
            $this->_generateUpdateScripts();
        }
        catch(Exception $e)
        {

            throw new Exception('Error setting up new  clone repository for staging area. '.$e->getMessage());
        }
    }

    function _makeStageRepositoryBackup($cloneDir)
    {

        $backupDir = get_dc_stage_repo_backup_path();
        $backupDirName =  $backupDir.DIRECTORY_SEPARATOR.date('Y-m-d-H-i');
        log_message('debug', 'Backing up staging area to new location ::'.$backupDirName);
        if (!@file_exists($backupDir))
        {
            $oldumask = umask(0);
            $result = @mkdir($backupDir,'0755');
            umask($oldumask);
            if (!$result)
            {
                log_message('debug', sprintf('Error creating backup directory to %s ::',$backupDirName));
                throw new Exception('Error creating backup directory for staging repository in location '.$backupDirName);
            }
        }
        if (@file_exists($cloneDir))
        {
            $renameResult = rename($cloneDir, $backupDirName);
            if (!$renameResult)
            {
               log_message('debug', sprintf('Backing up Failed from directory  %s to %s ::',$cloneDir,$backupDirName));
            }
        }
    }

    function gitInitStagingFromPath($path,$username)
    {
        log_message('debug', "Initializing git clone for staging area from given path: $path");
        $cloneDir = get_dc_staging_repository();
        try
        {
            if (empty($path))
            {
                throw new Error("Path cannot be empty");
            }
            if (is_dir($path) && !is_dir_empty($path))
            {
                @copyr($path,$cloneDir);
                // initialize a git repo here
                deleteAll($cloneDir.DIRECTORY_SEPARATOR.'.git',false);
                $repo = new CF_GitRepo($cloneDir, true, true,$username);
                $repo->add();
                $repo->commit("Initializing staging repository",'admin','admin@cfengine.com','admin@cfengine.com');
                $repo->create_branch(get_dc_git_working_branch());
                delete_git_update_scripts();
            }
            else
            {
                throw new Exception("Path supplied is either not a directory or is empty.");
            }
        }
        catch(Exception $e)
        {

            throw new Exception('Error setting up new  clone repository for staging area.'.$e->getMessage());
        }
    }

    function _generateUpdateScripts()
    {
        $this->_generateScriptFromTemplates('params.template','params.sh');
    }

    function _generateScriptFromTemplates($templateName,$filename)
    {

        $scriptTemplatePath =  get_dc_update_script_template_path().DIRECTORY_SEPARATOR.$templateName;
        $templateContent = @file_get_contents($scriptTemplatePath);
        if ($templateContent === false)
        {
            log_message('ERROR',"Template for $templateName script not found at path:".$scriptTemplatePath);
            throw new Exception("Cannot get the content of the $templateName script template to generate the script from path $scriptTemplatePath");
        }
        else
        {
            $replace = $this->_getScriptTemplateParameters();
            $generatedScript = str_replace(array_keys($replace), array_values($replace), $templateContent);
            $scriptPath = get_dc_update_script_path().DIRECTORY_SEPARATOR.$filename;
            log_message('debug',"writing $templateName script at path: $scriptPath");

            $result = @file_put_contents($scriptPath, $generatedScript);
            if ($result === false)
            {
                log_message('error',"writing $filename script at path: $scriptPath");
                throw new Exception("Cannot generate $filename script file at $scriptPath");
            } else
            {
                chmod($scriptPath, 0750);
            }
        }

    }

    function _getScriptTemplateParameters()
    {
        $gitServerUrl = $this->getGitServer();
        $gitBranch = $this->getGitBranch();
        $keyPath = get_dc_user_ssh_key();
        $workDir = get_dc_staging_repository();
        $gitLocalWorkingBranch = get_dc_git_working_branch();
        $scriptDir = get_dc_update_script_template_path();

        $replace = array(
            '%WORKING_DIR%' =>$workDir,
            '%REMOTE_GIT_URL%'=>$gitServerUrl,
            '%REMOTE_GIT_BRANCH%'=>$gitBranch,
            '%LOCAL_GIT_BRANCH%'=> $gitLocalWorkingBranch,
            '%SSH_PRIVATE_KEY%' => $keyPath,
            '%VCS_TYPE%' => 'GIT',
            '%SCRIPT_DIR%' => $scriptDir);
        return $replace;
    }

    function getExistingGitRepo()
    {

        $username = $this->getUsername();
        $repoPath = get_dc_user_working_directory($username);
        $repo = new CF_GitRepo($repoPath, false, false, $username);
        log_message('debug',"Getting git repo on path $repoPath");
        return $repo;
    }


    function checkStagingRepository()
    {

        $username = $this->getUsername();
        $repoPath = get_dc_staging_repository();
        log_message('debug',"Checking for staging repository at ::  $repoPath");
        try
        {
            $repo = new CF_GitRepo($repoPath, false, false, $username);
        }
        catch(Exception $e)
        {
            throw new DC_Staging_Notfound_Exception("Staging repository not found");
        }
        return $repo;
    }

    function checkUserRepository()
    {

        $username = $this->getUsername();
        $gitServerUrl = $this->getGitServer();
        $workDir = get_dc_user_working_directory($this->getUsername());
        $gitBranch = $this->getGitBranch();
        $keyPath = get_dc_user_ssh_key_directory($this->getUsername());

        log_message('debug',"Checking for user repository at :: $workDir");
        try
        {
            $repo = new CF_GitRepo($workDir, false, false, $username);
            $repo->reset_hard(get_dc_git_working_branch());
        }
        catch(Exception $e)
        {
           // no repository found.Initialize one
            try
            {
              @deleteAll($workDir);
              $repo = new CF_GitRepo($workDir, true, false, $username);
              $repo->clone_remote($gitServerUrl);
              $repo->checkout_remote_branch(get_dc_git_working_branch());
            }
            catch(Exception $e)
            {
                    $message = sprintf("Failed cloning repository with the url:: %s.
                                        Please make sure you have the Git server url (ssh).", $gitServerUrl);
                    log_message('error', 'Failed cloning repository :: ' . $e->getMessage());
                    throw new Exception($message);
            }
        }
    }

    function resetGitRepository()
    {
        $repo = $this->getExistingGitRepo();
        $repo->reset_hard($this->getGitBranch());
    }

    function commit_push($msg)
    {

        $repo = $this->getExistingGitRepo();
        $repo->add();
        $repo->commit($msg,$this->getGitAuthorName(), $this->getGitAuthorEmail(),$this->getUserEmail());
        $gitUrl = $this->getGitServer();
        $branch = $this->getGitBranch();
        $repo->push('origin', $branch);
        $repo->post_push();
    }

    function reset_and_pull()
    {
       $repo = $this->getExistingGitRepo();
       $repo->pre_fetch();
       $repo->fetch();
       $repo->reset_hard($this->getGitBranch());
    }

}

class DC_Exception extends Exception { }
class DC_Settings_Exception extends DC_Exception{}; // when settings are not specified
class DC_Staging_Notfound_Exception extends DC_Exception{};
class Git_Settings_Exception extends DC_Exception{}; // related to git errors
?>