
-------------------------------------------------------------------------------
-- Custom types:
-------------------------------------------------------------------------------

DO $$
BEGIN

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'change_operation') THEN
CREATE TYPE CHANGE_OPERATION AS ENUM ('ADD', 'CHANGE', 'REMOVE', 'UNTRACKED');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'agent_exec_status') THEN
CREATE TYPE AGENT_EXEC_STATUS AS ENUM ('OK', 'FAIL');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'patch_status') THEN
CREATE TYPE PATCH_STATUS AS ENUM ('INSTALLED', 'AVAILABLE');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'last_seen_direction') THEN
CREATE TYPE LAST_SEEN_DIRECTION AS ENUM ('INCOMING', 'OUTGOING');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'promise_outcome') THEN
CREATE TYPE PROMISE_OUTCOME AS ENUM ('KEPT', 'NOTKEPT', 'REPAIRED');
END IF;
END$$;
-------------------------------------------------------------------------------
-- Functions and Operators:
----------------------s---------------------------------------------------------

CREATE OR REPLACE FUNCTION reverse_lower_eq(text,text) 
RETURNS BOOL AS 
    'select lower($2) = lower($1)' 
LANGUAGE SQL IMMUTABLE STRICT PARALLEL SAFE;

-------------------------------------------------------------------------------

DROP OPERATOR IF EXISTS ~~~% (text, text);
CREATE OPERATOR ~~~% (procedure = reverse_lower_eq, leftarg = text, rightarg = text);

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION reverse_lower_like(text,text) 
RETURNS BOOL AS 
    'select lower($2) LIKE lower($1)' 
LANGUAGE SQL IMMUTABLE STRICT PARALLEL SAFE; 

-------------------------------------------------------------------------------

DROP OPERATOR IF EXISTS ~~~%~ (text, text);
CREATE OPERATOR ~~~%~ (procedure = reverse_lower_like, leftarg = text, rightarg = text);

-------------------------------------------------------------------------------

-- Functions for cfdb database used by Mission Portal

-- remove '{ ',' '} and {" "," "} from inventory slist fields
CREATE OR REPLACE FUNCTION cf_clearSlist(varchar) RETURNS varchar AS
$$ 	SELECT trim(regexp_replace(regexp_replace($1,'\'',\''|\'', \''|", ?"', ', ', 'g'), '(^\s*{\''|\''}|{"|"}\s*$)', '', 'g'))
$$ LANGUAGE SQL IMMUTABLE RETURNS NULL on NULL INPUT;

-- convert string to number, return NULL if not able to convert
-- Using EXCEPTION creates subtransaction that is not PARALLEL SAFE
CREATE OR REPLACE FUNCTION cf_convertToNum(text) RETURNS numeric AS $$
    DECLARE x NUMERIC;
    BEGIN
        x = $1::NUMERIC;
        RETURN x;
    EXCEPTION WHEN others THEN
        RETURN NULL;
    END;
$$ LANGUAGE plpgsql IMMUTABLE RETURNS NULL on NULL INPUT;


-- return hostkeys and grouped attributes values as table by attribute name
CREATE OR REPLACE FUNCTION getHostkeyAndValuesForAttribute(attrSubStr TEXT) RETURNS TABLE (
   hostkey   text,
   inventoryValue text
  ) AS
$func$

DECLARE sqlStr TEXT;

BEGIN

sqlStr := $x$
   SELECT HostKey,
	  string_agg(level2.value, ', ' ORDER BY lower(level2.value) ASC )
	  FROM
		   (SELECT
			inventory.HostKey,
			inventory.value
		   FROM
		      inventory
		   WHERE
			 $x$|| quote_nullable(attrSubStr)  ||$x$  = ANY (inventory.metatags)
		   GROUP BY
			inventory.hostkey,
			inventory.value) as level2
	GROUP BY HostKey;
    $x$;
RETURN query
EXECUTE sqlStr
USING   attrSubStr;

  end;  $func$ language plpgsql PARALLEL SAFE;


-- return hostkeys and grouped attributes values as table by key name

CREATE OR REPLACE FUNCTION getHostkeyAndValuesForKeyName(keyName TEXT) RETURNS TABLE (
   hostkey   text,
   inventoryValue text
  ) AS
$func$

DECLARE sqlStr TEXT;

BEGIN

sqlStr := $x$
   SELECT HostKey,
	  string_agg(level2.value, ', ' ORDER BY lower(level2.value) ASC )
	  FROM
		   (SELECT
			inventory.HostKey,
			inventory.value
		   FROM
		      inventory
		   WHERE
			 keyname = $x$  || quote_nullable(keyName) || $x$
		   GROUP BY
			inventory.hostkey,
			inventory.value) as level2
	GROUP BY HostKey;
    $x$;
RETURN query
EXECUTE sqlStr
USING   keyName;

  end;  $func$ language plpgsql PARALLEL SAFE;

-- generate partition name for promiselog
CREATE OR REPLACE FUNCTION promise_log_partition_name_create(date TIMESTAMPTZ, outcome PROMISE_OUTCOME)
RETURNS TEXT  AS 
$BODY$
DECLARE
    name TEXT;

BEGIN
    name := '__promiselog_' || outcome || '_' || to_char(date, 'YYYY-MM-DD');
    RETURN name;
END;
$BODY$
LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- Auto paritioning function for promise_log. 
-- Not generic at this point as it creates indexing as well
CREATE OR REPLACE FUNCTION promise_log_partition_function()
RETURNS TRIGGER AS 
$BODY$
DECLARE
    _tablename text;
    _startdate text;
    _enddate text;
    _outcome PROMISE_OUTCOME;
    _result record;
    person_exists boolean;

BEGIN
    -- Generate partition name based on ChangeTimeStamp and PromiseOutcome
    _outcome := NEW."promiseoutcome";
    _tablename := promise_log_partition_name_create(NEW."changetimestamp"::TIMESTAMPTZ, _outcome::PROMISE_OUTCOME);

    -- Insert the current record into the correct partition, which we are sure will now exist.
    EXECUTE 'INSERT INTO ' || quote_ident(_tablename) || ' VALUES ($1.*)' USING NEW;
    RETURN NULL;

    EXCEPTION
    WHEN undefined_table THEN

    PERFORM promise_log_partition_create(NEW."changetimestamp", 1, _outcome);
    EXECUTE 'INSERT INTO ' || quote_ident(_tablename) || ' VALUES ($1.*)' USING NEW;

    RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION promise_log_partition_create(now TIMESTAMPTZ, partitions INTEGER, outcome PROMISE_OUTCOME)
RETURNS VOID AS 
$BODY$

DECLARE
    _interval INTERVAL;
    _now TIMESTAMPTZ;
    _table_time TIMESTAMPTZ;
    _tablename TEXT;

BEGIN
    _interval := quote_literal(partitions - 1|| ' day');
    _now := to_timestamp(to_char(now, 'YYYY-MM-DD'), 'YYYY-MM-DD');

    FOR _table_time IN
        SELECT ts::timestamp 
            FROM generate_series(_now::timestamp, _now::timestamp + _interval, '1 day'::INTERVAL) as ts
    LOOP
        _tablename := promise_log_partition_name_create(_table_time, outcome);

        IF NOT EXISTS (
            SELECT 1
            FROM   information_schema.tables 
            WHERE  table_name = _tablename )
        THEN
            -- create new partition
            EXECUTE 'CREATE TABLE ' || quote_ident(_tablename) || ' (
                CHECK ( "changetimestamp" >= ' || quote_literal(_table_time) || '::timestamptz 
                AND "changetimestamp" < ' || quote_literal(_table_time + INTERVAL '1 day') || '::timestamptz 
                AND "promiseoutcome" = ' || quote_literal(outcome) || '::PROMISE_OUTCOME )
                ) INHERITS (__PromiseLog)';

            -- Indexes are defined per child, so we assign a default index that uses the partition columns
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_hostkey') || ' ON ' || quote_ident(_tablename) || ' (HostKey) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE UNIQUE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_id') || ' ON ' || quote_ident(_tablename) || ' (Id) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_outcome') || ' ON ' || quote_ident(_tablename) || ' (PromiseOutcome) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_changetimestamp') || ' ON ' || quote_ident(_tablename) || ' (ChangeTimeStamp) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_policyfile') || ' ON ' || quote_ident(_tablename) || ' (lower(PolicyFile)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_promisehash') || ' ON ' || quote_ident(_tablename) || ' (lower(PromiseHash)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_bundlename') || ' ON ' || quote_ident(_tablename) || ' (lower(BundleName)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_promisetype') || ' ON ' || quote_ident(_tablename) || ' (lower(PromiseType)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_promiser') || ' ON ' || quote_ident(_tablename) || ' (lower(Promiser)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_stackpath') || ' ON ' || quote_ident(_tablename) || ' (lower(StackPath)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_promisehandle') || ' ON ' || quote_ident(_tablename) || ' (lower(PromiseHandle)) WITH (FILLFACTOR = 90)';

        END IF;
    END LOOP;

    RETURN;
END;
$BODY$
LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- Cleanup promiselog based on interval
CREATE OR REPLACE FUNCTION promise_log_partition_cleanup(outcome PROMISE_OUTCOME, keep_days INTERVAL)
RETURNS VOID AS 
$BODY$
DECLARE
    _tablename text;

BEGIN
    FOR _tablename IN
        SELECT table_name 
        FROM (
            SELECT table_name, to_timestamp(substring(table_name from (15 + char_length(outcome::text)) for 10), 'YYYY MM DD') as timestamp
            FROM information_schema.tables
            WHERE table_schema='public'
                AND table_type='BASE TABLE'
                AND table_name ILIKE '__promiselog_' || outcome || '_%'
        ) as sub
        WHERE timestamp < NOW() - keep_days
    LOOP
        EXECUTE 'DROP TABLE ' || quote_ident(_tablename);
    END LOOP;

    RETURN;
END;
$BODY$
LANGUAGE plpgsql;

-- Create function to add new custom variables to variables_dictionary
CREATE OR REPLACE FUNCTION update_variables_dictionary()
    RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
       RAISE LOG 'PostgreSQL is in recovery, skipping update of variables dictionary';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping update of variables dictionary';
  ELSE
    INSERT INTO variables_dictionary ("attribute_name", "category","type","readonly","keyname")
        SELECT distinct on (attribute_name) "substring"(__variables.metatags::text, 'attribute_name=["]?([^"},]+)'::text) AS attribute_name,
        'User defined'::text as category,
        variabletype as type,
        '0'::int as readonly,
        comp as keyname
        FROM __variables
        WHERE  ('inventory'::text = ANY (__variables.metatags)) AND NOT ('attribute_name=none'::text = ANY (__variables.metatags))
        AND "substring"(__variables.metatags::text, 'attribute_name=["]?([^"},]+)'::text) NOT IN (SELECT attribute_name FROM variables_dictionary )
        GROUP BY __variables.metatags , variabletype, comp;

    INSERT INTO variables_dictionary ("attribute_name", "category","type","readonly","keyname")
        SELECT distinct on (attribute_name) "substring"(__contexts.metatags::text, 'attribute_name=["]?([^"},]+)'::text) AS attribute_name,
        'User defined'::text as category,
        'string'::text as type,
        '0'::int as readonly,
        ''::text as keyname
        FROM __contexts
        WHERE  ('inventory'::text = ANY (__contexts.metatags)) AND NOT ('attribute_name=none'::text = ANY (__contexts.metatags))
        AND "substring"(__contexts.metatags::text, 'attribute_name=["]?([^"},]+)'::text) NOT IN (SELECT attribute_name FROM variables_dictionary )
        GROUP BY __contexts.metatags;
  END IF;
END;
$$ LANGUAGE plpgsql;

-- Disable variable which don't have any reports data
CREATE OR REPLACE FUNCTION disable_unreported_variables()
  RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping disabling unreported variables';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping disabling unreported variables';
  ELSE
    UPDATE variables_dictionary SET enabled = 0;
    UPDATE variables_dictionary SET enabled = 1 WHERE attribute_name IN ( SELECT distinct on (attribute_name) "substring"(__variables.metatags::text, 'attribute_name=["]?([^"},]+)'::text) AS attribute_name  FROM __variables) OR attribute_name IN ( SELECT distinct on (attribute_name) "substring"(__contexts.metatags::text, 'attribute_name=["]?([^"},]+)'::text) AS attribute_name  FROM __contexts);
  END IF;
END;
$$ LANGUAGE plpgsql;

-- Create function  update_materialized_inventory_view
CREATE OR REPLACE FUNCTION update_materialized_inventory_view()
    RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping refresh of materialized view m_inventory';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping refresh of materialized view m_inventory';
  ELSE
    RAISE NOTICE 'Refreshing materialized view m_inventory';
        REFRESH MATERIALIZED VIEW CONCURRENTLY m_inventory;
  END IF;
END;
$$ LANGUAGE plpgsql;

-- Create function  update_materialized_hosts_view
CREATE OR REPLACE FUNCTION update_materialized_hosts_view()
    RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping refresh of materialized view m_hosts';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping refresh of materialized view m_hosts';
  ELSE
    RAISE NOTICE 'Refreshing materialized view m_hosts';
    REFRESH MATERIALIZED VIEW CONCURRENTLY m_hosts;
  END if;
END;
$$ LANGUAGE plpgsql;

-- Create function  ensure_view_with_rbac_filter
CREATE OR REPLACE FUNCTION ensure_view_with_rbac_filter(table_name TEXT, view_name TEXT, hostkey_column TEXT default 'HostKey') RETURNS void AS
$BODY$
BEGIN
  EXECUTE 'CREATE OR REPLACE VIEW ' || $2 || ' WITH (security_barrier) AS
    SELECT p.*
    FROM ' || $1 || ' p
    WHERE EXISTS (
            SELECT HostKey
            FROM ContextCache
            WHERE ContextVector @@ to_tsquery(''simple'', current_setting(''rbac.filter''::text))
                AND p.' || $3 || ' = HostKey
        );';
  RETURN;
END;
$BODY$ LANGUAGE plpgsql STRICT;

-- Create function  clean_historical_data
CREATE OR REPLACE FUNCTION cleanup_historical_data(
    table_name TEXT,
    order_by_column TEXT,
    items_number_to_keep INTEGER,
    hostkey_column TEXT default 'HostKey'
    ) RETURNS void AS
$BODY$
DECLARE
    row record;
BEGIN
    table_name := $1;
    order_by_column := $2;
    items_number_to_keep := $3;
    hostkey_column := $4;

    FOR row IN
        EXECUTE 'SELECT DISTINCT '|| hostkey_column ||' FROM '|| table_name ||''
    LOOP
        EXECUTE 'DELETE FROM '|| table_name ||' WHERE '|| order_by_column ||' NOT IN (SELECT '|| order_by_column ||' FROM '|| table_name ||' WHERE '|| hostkey_column ||' = '''|| row.host ||''' ORDER BY '|| order_by_column ||' DESC OFFSET 0 LIMIT '|| items_number_to_keep ||') AND '|| hostkey_column ||' = '''|| row.host ||'''';
    END LOOP;

  RETURN;
END;
$BODY$ LANGUAGE plpgsql STRICT;

COMMENT ON FUNCTION cleanup_historical_data (text, text, integer, text) IS 'Clean up stale records from table. Arguments: table name, order by column, number items to preserve, host key column name.';
-------------------------------------------------------------------------------
-- RBAC
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS ContextCache (
        HostKey             TEXT        NOT NULL,
        ContextVector       TSVECTOR    NOT NULL
    ) WITH (FILLFACTOR = 90);

CREATE UNIQUE INDEX IF NOT EXISTS contexts_cache_uniq ON ContextCache (HostKey) WITH (FILLFACTOR = 90);
CREATE INDEX IF NOT EXISTS context_cache_vector ON ContextCache USING gin(ContextVector);


-------------------------------------------------------------------------------
-- Current state tables:
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __Hosts (
        HostKey                 TEXT                        NOT NULL,
        IsCallCollected         boolean                     DEFAULT false,
        LastReportTimeStamp     timestamp with time zone    DEFAULT CURRENT_TIMESTAMP,
--      HostName                TEXT,                       comment: virtual field in hosts view mapped from variables
--      IPAddress               TEXT,                       comment: virtual field in hosts view mapped from lastseenhosts
        FirstReportTimeStamp    timestamp with time zone    DEFAULT CURRENT_TIMESTAMP
    ) WITH (FILLFACTOR = 70);

CREATE UNIQUE INDEX IF NOT EXISTS hosts_uniq ON __Hosts (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__Hosts', 'v_Hosts');

CREATE OR REPLACE VIEW hosts_not_reported WITH (security_barrier) AS
    SELECT p.*
    FROM __hosts p
    WHERE EXISTS (
            SELECT HostKey
            FROM ContextCache
            WHERE ContextVector @@ to_tsquery('simple', 'host.not.reported')
                AND p.hostkey = HostKey
        );

COMMENT ON VIEW hosts_not_reported IS 'View to select never collected hosts';

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __AgentStatus (
        HostKey                     		TEXT                          NOT NULL,
        AgentExecutionInterval      		INTEGER,
        LastAgentLocalExecutionTimeStamp 	timestamp with time zone,
        LastAgentExecutionStatus    		AGENT_EXEC_STATUS
    ) WITH (FILLFACTOR = 70);

CREATE UNIQUE INDEX IF NOT EXISTS agentstatus_uniq ON __AgentStatus (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__AgentStatus', 'AgentStatus');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __Contexts (
        HostKey             TEXT                        NOT NULL,
        ContextName         TEXT                        NOT NULL,
        MetaTags            TEXT[],
        ChangeTimeStamp     timestamp with time zone
    ) WITH (FILLFACTOR = 90);

CREATE UNIQUE INDEX IF NOT EXISTS contexts_uniq ON __Contexts (HostKey, ContextName) WITH (FILLFACTOR = 90);
CREATE INDEX IF NOT EXISTS contexts_pattern_ops ON __Contexts (ContextName varchar_pattern_ops) WITH (FILLFACTOR = 90);
CREATE INDEX IF NOT EXISTS contexts_meta ON __Contexts (MetaTags) WITH (FILLFACTOR = 90);

SELECT ensure_view_with_rbac_filter('__Contexts', 'Contexts');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __Variables (
        HostKey             TEXT                        NOT NULL,
        NameSpace           TEXT                        DEFAULT NULL,
        Bundle              TEXT                        NOT NULL,
        VariableName        TEXT                        NOT NULL,
        VariableValue       TEXT                        DEFAULT NULL,
        VariableType        TEXT                        NOT NULL,
        comp                TEXT                        NOT NULL,
        MetaTags            TEXT[],
        ChangeTimeStamp     timestamp with time zone
    ) WITH (FILLFACTOR = 70);

CREATE UNIQUE INDEX IF NOT EXISTS variables_uniq ON __Variables (HostKey, comp) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS variables_meta ON __Variables (MetaTags) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS variables_variablename ON __Variables (VariableName) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS variables_variablevalue ON __Variables (VariableValue varchar_pattern_ops) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__Variables', 'Variables');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __Software (
        HostKey                 TEXT                        NOT NULL,
        compo                   TEXT                        NOT NULL,
        SoftwareName            TEXT                        NOT NULL,
        SoftwareVersion         TEXT                        NOT NULL,
        SoftwareArchitecture    TEXT                        NOT NULL,
        ChangeTimeStamp		    timestamp with time zone
    ) WITH (FILLFACTOR = 90);

CREATE UNIQUE INDEX IF NOT EXISTS software_uniq ON __Software (HostKey, compo) WITH (FILLFACTOR = 90);
CREATE INDEX IF NOT EXISTS software_name_version ON __Software (SoftwareName, SoftwareVersion) with (FILLFACTOR = 90);

SELECT ensure_view_with_rbac_filter('__Software', 'Software');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __SoftwareUpdates (
        HostKey             TEXT                        NOT NULL,
        compo               TEXT                        NOT NULL,
        PatchName           TEXT                        NOT NULL,
        PatchVersion        TEXT                        NOT NULL,
        PatchArchitecture   TEXT                        NOT NULL,
        PatchReportType     PATCH_STATUS                NOT NULL,
        ChangeTimeStamp   	timestamp with time zone
    ) WITH (FILLFACTOR = 90);

CREATE UNIQUE INDEX IF NOT EXISTS softwareupdates_uniq ON __SoftwareUpdates (HostKey, compo) WITH (FILLFACTOR = 90);
CREATE INDEX IF NOT EXISTS software_name ON __SoftwareUpdates (PatchName) WITH (FILLFACTOR = 90);

SELECT ensure_view_with_rbac_filter('__SoftwareUpdates', 'SoftwareUpdates');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __LastSeenHosts (
        HostKey             TEXT                        NOT NULL,
        compo               TEXT                        NOT NULL,
        LastSeenDirection   LAST_SEEN_DIRECTION         NOT NULL,
        RemoteHostKey       TEXT                        NOT NULL,
        RemoteHostIP        TEXT,
        LastSeenTimeStamp   timestamp with time zone    NOT NULL,
        LastSeenInterval    real                        NOT NULL
    ) WITH (FILLFACTOR = 70);

CREATE UNIQUE INDEX IF NOT EXISTS lastseenhosts_uniq ON __LastSeenHosts (HostKey, compo) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS lastseenhosts_remotekey ON __LastSeenHosts (RemoteHostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__LastSeenHosts', 'LastSeenHosts');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __LastSeenHostsLogs (
        Id                  BIGSERIAL                   PRIMARY KEY,
        HostKey             TEXT                        NOT NULL,
        compo               TEXT                        NOT NULL,
        LastSeenDirection   LAST_SEEN_DIRECTION         NOT NULL,
        RemoteHostKey       TEXT                        NOT NULL,
        RemoteHostIP        TEXT,
        LastSeenTimeStamp   timestamp with time zone    NOT NULL,
        LastSeenInterval    real                        NOT NULL
    ) WITH (FILLFACTOR = 70);

------------- Drop fk_lastSeenHostsLogs_hostkey added in 3.12--------------------------
ALTER TABLE __LastSeenHostsLogs DROP CONSTRAINT IF EXISTS fk_lastSeenHostsLogs_hostkey;

CREATE INDEX IF NOT EXISTS lastseenhostlogs_hostkey ON __LastSeenHostsLogs (HostKey) WITH (FILLFACTOR = 70);

-- Drop view to change column type to bigint --
DROP VIEW IF EXISTS LastSeenHostsLogs;
ALTER TABLE __LastSeenHostsLogs ALTER COLUMN id type bigint;

SELECT ensure_view_with_rbac_filter('__LastSeenHostsLogs', 'LastSeenHostsLogs');

CREATE OR REPLACE FUNCTION lastseen_hosts_log_function()
RETURNS TRIGGER AS
$BODY$
BEGIN
       IF (NEW.lastseendirection = 'INCOMING') THEN
             INSERT INTO "__lastseenhostslogs"  (HostKey, compo,  LastSeenDirection, RemoteHostKey, RemoteHostIP, LastSeenTimeStamp, LastSeenInterval   ) VALUES (NEW.*)   ;
       END IF;
    RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS lastseen_hosts_insert_trigger ON __LastSeenHosts;
CREATE TRIGGER lastseen_hosts_insert_trigger
AFTER INSERT OR UPDATE ON __LastSeenHosts
FOR EACH ROW EXECUTE PROCEDURE lastseen_hosts_log_function();



-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __PromiseExecutions (
        HostKey             TEXT                        NOT NULL,
        PolicyFile          TEXT                        NOT NULL,
        ReleaseId           TEXT                        DEFAULT NULL,
        PromiseHash         TEXT                        NOT NULL,
        NameSpace           TEXT                        NOT NULL,
        BundleName          TEXT                        NOT NULL,
        PromiseType         TEXT                        NOT NULL,
        Promiser            TEXT                        DEFAULT NULL,
        StackPath           TEXT                        NOT NULL,
        PromiseHandle       TEXT                        DEFAULT NULL,
        PromiseOutcome      CHARACTER VARYING(8)        NOT NULL,
        LogMessages         TEXT[]                      DEFAULT NULL,
        Promisees           TEXT[]                      DEFAULT NULL,
        ChangeTimeStamp     timestamp with time zone
    ) WITH (FILLFACTOR = 70);

-- change promiseoutcome data type to character varying
DO $$
  BEGIN
    IF NOT EXISTS (
        SELECT 1 from information_schema.columns
        WHERE table_name ILIKE '__promiseexecutions'
          AND table_schema = 'public'
          AND column_name = 'promiseoutcome'
          AND data_type = 'character varying')
    THEN
      DROP VIEW IF EXISTS promiseexecutions;
      ALTER TABLE __PromiseExecutions ALTER COLUMN promiseoutcome type CHARACTER VARYING(8) USING promiseoutcome::text;
    END IF;
END$$;

CREATE UNIQUE INDEX IF NOT EXISTS promiseexecutions_uniq ON __PromiseExecutions (HostKey, PromiseHash) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS promiseexecutions_bundlename ON __PromiseExecutions (BundleName) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS promiseexecutions_promiser ON __PromiseExecutions (Promiser) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS promiseexecutions_promiseoutsome ON __PromiseExecutions (PromiseOutcome) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS promiseexecutions_promisees ON __PromiseExecutions (Promisees) WITH (FILLFACTOR =  70);
CREATE INDEX IF NOT EXISTS promiseexecutions_promisehash ON __PromiseExecutions (PromiseHash) WITH (FILLFACTOR =  70);

SELECT ensure_view_with_rbac_filter('__PromiseExecutions', 'PromiseExecutions');

-------------------------------------------------------------------------------

-- Virtual View
--CREATE TABLE IF NOT EXISTS Inventory (
--        HostKey             TEXT,
--        KeyName             TEXT,
--        Type                TEXT,
--        MetaTags            TEXT[],
--        Value               TEXT
--    );

CREATE OR REPLACE VIEW Inventory WITH (security_barrier) AS (
    SELECT 
        HostKey,
        KeyName,
        Type,
        MetaTags,
        Value 
    FROM (
        SELECT 
            HostKey, 
            ContextName AS KeyName,
            'context' AS Type,
            MetaTags,
            ContextName AS Value
        FROM 
            Contexts
        WHERE 
            'inventory' = ANY (MetaTags)
            AND NOT 'attribute_name=none' = ANY(MetaTags)
        UNION ALL
        SELECT 
            HostKey,
            comp AS KeyName,
            VariableType AS Type,
            MetaTags,
            VariableValue AS Value
        FROM 
            Variables
        WHERE 
            'inventory' = ANY (MetaTags) 
            AND NOT 'attribute_name=none' = ANY(MetaTags)
    ) tmp 
);


CREATE OR REPLACE VIEW Hosts WITH (security_barrier) AS (
    SELECT
        v_Hosts.hostkey,
        v.HostName,
        l.IPAddress,
        v_Hosts.LastReportTimeStamp,
        v_Hosts.FirstReportTimeStamp
    FROM
        v_Hosts
    LEFT JOIN (
        SELECT
            HostKey,
            VariableValue AS HostName
        FROM __Variables
        WHERE comp=current_setting('host.identifier'::text)
    ) v
    ON v_Hosts.HostKey=v.HostKey
    LEFT JOIN (
        SELECT
            HostKey,
            IPAddress
        FROM (
            SELECT
                RemoteHostKey AS HostKey,
                RemoteHostIP AS IPAddress,
                ROW_NUMBER() OVER (PARTITION BY RemoteHostKey)
            FROM __LastSeenHosts
        ) AS CTE
        WHERE
            row_number=1
    ) l
    ON v_Hosts.HostKey=l.HostKey
);

-------------------------------------------------------------------------------
-- Log tables:
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __ContextsLog (
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeOperation     CHANGE_OPERATION            NOT NULL,
        ContextName         TEXT                        NOT NULL,
        MetaTags            TEXT[]
    );

CREATE INDEX IF NOT EXISTS contexts_log_hostkey ON __ContextsLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__ContextsLog', 'ContextsLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __VariablesLog (
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeOperation     CHANGE_OPERATION            NOT NULL,
        NameSpace           TEXT                        DEFAULT NULL,
        Bundle              TEXT                        NOT NULL,
        VariableName        TEXT                        NOT NULL,
        VariableValue       TEXT                        DEFAULT NULL,
        VariableType        TEXT                        NOT NULL,
        MetaTags            TEXT[]
    );

CREATE INDEX IF NOT EXISTS variables_log_hostkey ON __VariablesLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__VariablesLog', 'VariablesLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __SoftwareLog (
        HostKey                 TEXT                        NOT NULL,
        ChangeTimeStamp         timestamp with time zone    NOT NULL,
        ChangeOperation         CHANGE_OPERATION            NOT NULL,
        SoftwareName            TEXT                        NOT NULL,
        SoftwareVersion         TEXT                        NOT NULL,
        SoftwareArchitecture    TEXT                        NOT NULL
    );

CREATE INDEX IF NOT EXISTS software_log_hostkey ON __SoftwareLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__SoftwareLog', 'SoftwareLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __SoftwareUpdatesLog (
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeOperation     CHANGE_OPERATION            NOT NULL,
        PatchName           TEXT                        NOT NULL,
        PatchVersion        TEXT                        NOT NULL,
        PatchArchitecture   TEXT                        NOT NULL,
        PatchReportType     PATCH_STATUS                NOT NULL
    );

CREATE INDEX IF NOT EXISTS software_updates_log_hostkey ON __SoftwareUpdatesLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__SoftwareUpdatesLog', 'SoftwareUpdatesLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __FileChangesLog (
        HostKey             TEXT                        NOT NULL,
        PromiseHandle       TEXT                        NOT NULL,
        FileName            TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeType          TEXT                        NOT NULL,
        ChangeDetails       TEXT[]                      DEFAULT NULL
    );

CREATE INDEX IF NOT EXISTS filechanges_hk ON __FileChangesLog (HostKey) WITH (FILLFACTOR = 90);

SELECT ensure_view_with_rbac_filter('__FileChangesLog', 'FileChangesLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __BenchmarksLog (
        HostKey             TEXT                        NOT NULL,
        EventName           TEXT                        NOT NULL,
        StandardDeviation   NUMERIC                     NOT NULL,
        AverageValue        NUMERIC                     NOT NULL,
        LastValue           NUMERIC                     NOT NULL,
        CheckTimeStamp      timestamp with time zone    NOT NULL
    );

CREATE INDEX IF NOT EXISTS benchmarkslog_hk ON __BenchmarksLog (HostKey) WITH (FILLFACTOR = 90);

SELECT ensure_view_with_rbac_filter('__BenchmarksLog', 'BenchmarksLog');

-------------------------------------------------------------------------------

-- Deprecated in 3.7.0 (replaces by PromiseLog)
-- TODO: Remove in 3.8.0

CREATE TABLE IF NOT EXISTS __PromiseExecutionsLog (
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeOperation     CHANGE_OPERATION            NOT NULL,
        PolicyFile          TEXT                        NOT NULL,
        ReleaseId           TEXT                        DEFAULT NULL,
        PromiseHash         TEXT                        NOT NULL,
        NameSpace           TEXT                        NOT NULL,
        BundleName          TEXT                        NOT NULL,
        PromiseType         TEXT                        NOT NULL,
        Promiser            TEXT                        DEFAULT NULL,
        StackPath           TEXT                        NOT NULL,
        PromiseHandle       TEXT                        DEFAULT NULL,
        PromiseOutcome      PROMISE_OUTCOME             NOT NULL,
        LogMessages         TEXT[]                      DEFAULT NULL,
        Promisees           TEXT[]                      DEFAULT NULL
    );

CREATE INDEX IF NOT EXISTS promiseexecutions_log_hostkey ON __PromiseExecutionsLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__PromiseExecutionsLog', 'PromiseExecutionsLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __PromiseLog (
        Id                  bigserial                   primary key,
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        PolicyFile          TEXT                        NOT NULL,
        ReleaseId           TEXT                        DEFAULT NULL,
        PromiseHash         TEXT                        NOT NULL,
        NameSpace           TEXT                        NOT NULL,
        BundleName          TEXT                        NOT NULL,
        PromiseType         TEXT                        NOT NULL,
        Promiser            TEXT                        DEFAULT NULL,
        StackPath           TEXT                        NOT NULL,
        PromiseHandle       TEXT                        DEFAULT NULL,
        PromiseOutcome      PROMISE_OUTCOME             NOT NULL,
        LogMessages         TEXT[]                      DEFAULT NULL,
        Promisees           TEXT[]                      DEFAULT NULL
    );

CREATE INDEX IF NOT EXISTS promiselog_promiseoutcome ON __PromiseLog (PromiseOutcome);
CREATE INDEX IF NOT EXISTS promiselog_promisehash ON __PromiseLog (PromiseHash);
CREATE INDEX IF NOT EXISTS promiselog_changetimestamp_promiseoutcome ON __promiselog ("changetimestamp", "promiseoutcome");

-- Drop view to change column type to bigint --
DROP VIEW IF EXISTS PromiseLog;
ALTER TABLE __PromiseLog ALTER COLUMN id type bigint;

SELECT ensure_view_with_rbac_filter('__PromiseLog', 'PromiseLog');

DROP TRIGGER IF EXISTS promise_log_partition_trigger ON __PromiseLog;
CREATE TRIGGER promise_log_partition_trigger
BEFORE INSERT ON __PromiseLog
FOR EACH ROW EXECUTE PROCEDURE promise_log_partition_function();

-- Create empty partitions during installation for 7 days in the past and 3 days up in the future.
SELECT promise_log_partition_create(NOW() - INTERVAL '7 day', 7 + 3, 'REPAIRED');
SELECT promise_log_partition_create(NOW() - INTERVAL '7 day', 7 + 3, 'NOTKEPT');

-------------------------------------------------------------------------------
-- Monitoring tables:
-------------------------------------------------------------------------------

-- 3.6.2 changes schema for __MonitoringMg, during upgrade old one has to be removed
DO $$ 
DECLARE
TableExists boolean;
BEGIN
    SELECT INTO TableExists 1 FROM information_schema.tables WHERE table_name = lower('__MonitoringMgMeta');
    IF NOT FOUND THEN
        DROP TABLE IF EXISTS __MonitoringMg;
    END IF;
END
$$ ;


CREATE TABLE IF NOT EXISTS __MonitoringMgMeta (
        id                  bigserial   PRIMARY KEY,
        hostkey             TEXT        NOT NULL,
        observable          TEXT        NOT NULL,
        global              BOOLEAN,
        expected_min        REAL,
        expected_max        REAL,
        unit                TEXT,
        description         TEXT,
        updatedTimeStamp    timestamp with time zone,
        lastUpdatedSample   INTEGER
) WITH (FILLFACTOR = 70);

-- Drop view to change column type to bigint --
DROP VIEW IF EXISTS MonitoringMgMeta;
ALTER TABLE __MonitoringMgMeta ALTER COLUMN id type bigint;

CREATE UNIQUE INDEX IF NOT EXISTS monitoring_mg_meta
ON __MonitoringMgMeta (hostkey, observable) 
WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__MonitoringMgMeta', 'MonitoringMgMeta');

CREATE TABLE IF NOT EXISTS __MonitoringMg (
        meta_id                bigserial     REFERENCES __MonitoringMgMeta (id) ON DELETE CASCADE,
        sample                 INTEGER,
        value1                 REAL,
        value2                 REAL,
        value3                 REAL,
        value4                 REAL,
        updatedTimeStamp       timestamp with time zone
) WITH (FILLFACTOR = 70);

ALTER TABLE __MonitoringMg ALTER COLUMN meta_id type bigint;


CREATE UNIQUE INDEX IF NOT EXISTS monitoring_mg
ON __MonitoringMg (meta_id, sample) 
WITH (FILLFACTOR = 70);

CREATE INDEX IF NOT EXISTS monitoring_mg_time ON __MonitoringMg (updatedTimeStamp) WITH (FILLFACTOR = 70);

-------------------------------------------------------------------------------

-- 3.6.2 changes schema for __MonitoringYr, during upgrade old one has to be removed
DO $$ 
DECLARE
TableExists boolean;
BEGIN
    SELECT INTO TableExists 1 FROM information_schema.tables WHERE table_name = lower('__MonitoringYrMeta');
    IF NOT FOUND THEN
        DROP TABLE IF EXISTS __MonitoringYr;
    END IF;
END
$$ ;


CREATE TABLE IF NOT EXISTS __MonitoringYrMeta (
        id                  bigserial   PRIMARY KEY,
        hostkey             TEXT        NOT NULL,
        observable          TEXT        NOT NULL,
        global              BOOLEAN,
        expected_min        REAL,
        expected_max        REAL,
        unit                TEXT,
        description         TEXT,
        lastUpdatedSample   INTEGER
) WITH (FILLFACTOR = 70);

-- Drop view to change column type to bigint --
DROP VIEW IF EXISTS MonitoringYrMeta;
ALTER TABLE __MonitoringYrMeta ALTER COLUMN id type bigint;

CREATE UNIQUE INDEX IF NOT EXISTS monitoring_yr_meta
ON __MonitoringYrMeta (hostkey, observable) 
WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__MonitoringYrMeta', 'MonitoringYrMeta');


CREATE TABLE IF NOT EXISTS __MonitoringYr (
        meta_id                bigserial     REFERENCES __MonitoringYrMeta (id) ON DELETE CASCADE,
        sample                 INTEGER,
        value1                 REAL,
        value2                 REAL,
        value3                 REAL,
        value4                 REAL
) WITH (FILLFACTOR = 70);

ALTER TABLE __MonitoringYr ALTER COLUMN meta_id type bigint;

CREATE UNIQUE INDEX IF NOT EXISTS monitoring_yr
ON __MonitoringYr (meta_id, sample) 
WITH (FILLFACTOR = 70);

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __MonitoringHg (
        host    TEXT        NOT NULL,
        id      TEXT        NOT NULL,
        ar1     REAL[]
    ) WITH (FILLFACTOR = 70);

CREATE UNIQUE INDEX IF NOT EXISTS monitoringhg_hk_id ON __MonitoringHg (host, id) WITH (FILLFACTOR = 70);

SELECT ensure_view_with_rbac_filter('__MonitoringHg', 'MonitoringHg', 'host');

-------------------------------------------------------------------------------
-- Diagnostics tables:
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __HubConnectionErrors (
        CheckTimeStamp      timestamp with time zone    DEFAULT CURRENT_TIMESTAMP,
        HostKey             TEXT                        NOT NULL,
        Message             TEXT                        NOT NULL,
        QueryType           TEXT                        NOT NULL
    );

SELECT ensure_view_with_rbac_filter('__HubConnectionErrors', 'HubConnectionErrors');

-------------------------------------------------------------------------------

-- 3.6.2 Renames '__Diagnostics' to 'Diagnostics' table
ALTER TABLE IF EXISTS __Diagnostics RENAME TO Diagnostics; 


CREATE TABLE IF NOT EXISTS Diagnostics (
        Name           TEXT                         NOT NULL,
        Details        TEXT,
        TimeStamp      timestamp with time zone     DEFAULT CURRENT_TIMESTAMP,
        Value          NUMERIC                      NOT NULL,
        Units          TEXT                         NOT NULL
    );

-- 3.14.0/3.12.2 renames 'Status' to '__Status' and creates a RBAC-checked view
DO $$
BEGIN
IF EXISTS (SELECT table_name
           FROM information_schema.tables
           WHERE table_name = LOWER('Status')
           AND table_type = 'BASE TABLE') THEN
  ALTER TABLE Status RENAME TO __Status;
END IF;
END$$;

CREATE TABLE IF NOT EXISTS __STATUS (
        host           CHARACTER VARYING(256)   NOT NULL,
        ts             CHARACTER VARYING(256)   NOT NULL,
        status         CHARACTER VARYING(256),
        lstatus        CHARACTER VARYING(256),
        type           CHARACTER,
        who            INTEGER,
        whr            INTEGER,
        PRIMARY        KEY(host,ts)
    ) WITH (FILLFACTOR = 90);

COMMENT ON TABLE __STATUS IS 'Report collection status';
COMMENT ON COLUMN __STATUS.ts IS 'Timestamp of last data provided by client during report collection. This is used by delta queries to request a start time.';

SELECT ensure_view_with_rbac_filter('__Status', 'Status', 'host');


-------------------------------------------------------------------------------

-- Update ContextCache after all tables are created (Useful during hub upgrade)
-- Warning: order is important.

TRUNCATE TABLE ContextCache;

INSERT INTO ContextCache (hostkey, contextvector)
    SELECT hostkey, to_tsvector('simple', replace(x::text,'_','.')) 
    FROM (
        SELECT hostkey, array_agg(contextname) as x 
        FROM __Contexts
        GROUP BY hostkey
        ) as sub;

INSERT INTO ContextCache(hostkey, contextvector) 
    SELECT hostkey, to_tsvector('simple',replace('host_not_reported'::text,'_','.'))
    FROM __hosts 
    WHERE hostkey NOT IN (SELECT hostkey FROM ContextCache);

-- Create MATERIALIZED  VIEW  public.m_inventory
DROP MATERIALIZED VIEW IF EXISTS public.m_inventory;

CREATE MATERIALIZED  VIEW public.m_inventory AS
SELECT __hosts.hostkey,
( SELECT json_object(array_agg(d.keyname), array_agg(d.value)) AS array_to_json
  FROM (
    SELECT  keyname,
    string_agg(cf_clearslist(value),', ') AS value
    FROM( SELECT substring(__variables.metatags::TEXT FROM 'attribute_name=["]?([^"},]+)') AS keyname,
      string_agg(cf_clearslist(__variables.variablevalue),', ')   AS value

      FROM __variables
      WHERE __variables.hostkey = __hosts.hostkey
      AND ('inventory'::text = ANY (__variables.metatags)) AND (__variables.metatags::text LIKE '%attribute_name%') AND NOT ('attribute_name=none'::text = ANY (__variables.metatags))
      GROUP BY __variables.metatags

      UNION ALL

      SELECT substring(__contexts.metatags::TEXT FROM 'attribute_name=["]?([^"},]+)') AS keyname,
      string_agg(cf_clearslist(__contexts.contextname),', ') AS value
      FROM __contexts
      WHERE __contexts.hostkey = __hosts.hostkey
      AND ('inventory'::text = ANY (__contexts.metatags)) AND (__contexts.metatags::text LIKE '%attribute_name%') AND NOT ('attribute_name=none'::text = ANY (__contexts.metatags))
      GROUP BY __contexts.metatags)f  GROUP BY keyname
    )d
  ) AS "values"
FROM __hosts;

CREATE UNIQUE INDEX IF NOT EXISTS m_inventory_hostkey ON m_inventory (hostkey);

DROP MATERIALIZED VIEW IF EXISTS public.inventory_new;
SELECT ensure_view_with_rbac_filter('m_inventory', 'inventory_new');

-- Create MATERIALIZED  VIEW  public.m_hosts
DROP MATERIALIZED VIEW IF EXISTS public.m_hosts;
SET host.identifier = 'default.sys.fqhost';
CREATE MATERIALIZED VIEW public.m_hosts AS
SELECT
        __hosts.hostkey,
        v.HostName,
        l.IPAddress,
        __hosts.LastReportTimeStamp,
        __hosts.FirstReportTimeStamp
    FROM
        __hosts
    LEFT JOIN (
        SELECT
            HostKey,
            VariableValue AS HostName
        FROM __Variables
        WHERE comp=current_setting('host.identifier'::text)
    ) v
    ON __hosts.HostKey=v.HostKey
    LEFT JOIN (
        SELECT
            HostKey,
            IPAddress
        FROM (
            SELECT
                RemoteHostKey AS HostKey,
                RemoteHostIP AS IPAddress,
                ROW_NUMBER() OVER (PARTITION BY RemoteHostKey)
            FROM __LastSeenHosts
        ) AS CTE
        WHERE
            row_number=1
    ) l
    ON __hosts.HostKey=l.HostKey;

CREATE UNIQUE INDEX IF NOT EXISTS m_hosts_hostkey ON m_hosts (hostkey);
ALTER TABLE public.m_hosts OWNER TO cfapache;

SELECT ensure_view_with_rbac_filter('m_hosts', 'vm_hosts');

-- Refresh materialized view in case of a host was bootstrapped
CREATE OR REPLACE FUNCTION refresh_m_hosts_function()
RETURNS TRIGGER AS
$BODY$
BEGIN
  SET host.identifier = 'default.sys.fqhost';
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping bootstrap triggered refresh of materialized view m_hosts';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping bootstrap triggered refresh of materialized view m_hosts';
  ELSE
    RAISE NOTICE 'Host bootstrap triggering refreshing m_hosts';
    REFRESH MATERIALIZED VIEW CONCURRENTLY m_hosts;
  END IF;
  RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS hosts_insert_trigger ON __hosts;
CREATE TRIGGER hosts_insert_trigger
AFTER INSERT ON __hosts
FOR EACH ROW EXECUTE PROCEDURE refresh_m_hosts_function();

DROP TRIGGER IF EXISTS hosts_delete_trigger ON __hosts;
CREATE TRIGGER hosts_delete_trigger
AFTER DELETE ON __hosts
FOR EACH STATEMENT EXECUTE PROCEDURE refresh_m_hosts_function();

-- Clear hosts references in case of a host was deleted
CREATE OR REPLACE FUNCTION clear_hosts_references()
RETURNS TRIGGER AS
$BODY$
BEGIN
    DELETE FROM __variables WHERE hostkey = OLD.hostkey;
    DELETE FROM __contexts WHERE hostkey = OLD.hostkey;
    DELETE FROM __contextslog WHERE hostkey = OLD.hostkey;
    DELETE FROM __lastseenhosts WHERE remotehostkey = OLD.hostkey;
    DELETE FROM __lastseenhostslogs WHERE remotehostkey = OLD.hostkey;
    DELETE FROM __software WHERE hostkey = OLD.hostkey;
    DELETE FROM __softwarelog WHERE hostkey = OLD.hostkey;
    DELETE FROM __softwareupdates WHERE hostkey = OLD.hostkey;
    DELETE FROM __softwareupdateslog WHERE hostkey = OLD.hostkey;
    DELETE FROM __promiseexecutions WHERE hostkey = OLD.hostkey;
    DELETE FROM __monitoringyrmeta WHERE hostkey = OLD.hostkey;
    DELETE FROM __monitoringmgmeta WHERE hostkey = OLD.hostkey;
    DELETE FROM __filechangeslog WHERE hostkey = OLD.hostkey;
    DELETE FROM __agentstatus WHERE hostkey = OLD.hostkey;
    DELETE FROM __benchmarkslog WHERE hostkey = OLD.hostkey;
    RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS hosts_clear_refrences_trigger ON __hosts;
CREATE TRIGGER hosts_clear_refrences_trigger
AFTER DELETE ON __hosts
FOR EACH ROW EXECUTE PROCEDURE clear_hosts_references();

-- DROP VIEW public.v_contextcache
DROP VIEW IF EXISTS public.v_contextcache;

-- Create TABLE public.variables_dictionary
CREATE TABLE IF NOT EXISTS public.variables_dictionary
(
  id serial   PRIMARY KEY,
  attribute_name character varying(200),
  category character varying(200),
  readonly integer DEFAULT 0,
  type character varying(200),
  convert_function character varying(200),
  keyname character varying(200),
  CONSTRAINT variables_dictionary_attribute_name UNIQUE (attribute_name)
)
WITH (
  OIDS=FALSE
);

ALTER TABLE variables_dictionary
ADD COLUMN IF NOT EXISTS enabled integer DEFAULT 0;

ALTER TABLE public.variables_dictionary
  OWNER TO cfpostgres;
GRANT SELECT ON TABLE public.variables_dictionary TO public;
GRANT ALL ON TABLE public.variables_dictionary TO cfpostgres;


-------------------------------------------------------------------------------
-- Events table :
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS  events (
    id        bigserial   PRIMARY KEY,
    event character varying(30) NOT NULL,
    payload json DEFAULT '{}' NOT NULL,
    date timestamptz DEFAULT CURRENT_TIMESTAMP NOT NULL
) WITH (FILLFACTOR = 90);

-------------------------------------------------------------------------------
-- Function and trigger to handle decommissioned and bootstrapped hosts events :
-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION process_host_events() RETURNS TRIGGER AS $host_log_event_trigger$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping processing host events';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping processing host events';
  ELSE
    RAISE NOTICE 'Processing host events';
    IF (TG_OP = 'DELETE') THEN
      INSERT INTO events  (event, payload) VALUES ('host_decommissioned', json_build_object('hostkey', OLD.hostkey));
      RETURN OLD;
    ELSIF (TG_OP = 'INSERT') THEN
      INSERT INTO events  (event, payload) VALUES ('host_bootstrapped', json_build_object('hostkey', NEW.hostkey));
      RETURN NEW;
    END IF;
  END IF;
  RETURN NULL;
END;
$host_log_event_trigger$ LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS host_log_event_trigger ON __hosts;
CREATE TRIGGER host_log_event_trigger
AFTER INSERT OR DELETE ON __hosts
    FOR EACH ROW EXECUTE PROCEDURE process_host_events();

---- ENT-3950 ----

CREATE TABLE IF NOT EXISTS "__health_diagnostics_dismissed" (
    hostkey text,
    report_type character varying(30),
    username character varying(20),
    CONSTRAINT "health_diagnostics_dismissed_hostkey_report_type_username" PRIMARY KEY ("hostkey", "report_type", "username")
);

COMMENT ON TABLE "__health_diagnostics_dismissed" IS 'Table that contains specific dismissed hosts from Health diagnostic reports';

SELECT ensure_view_with_rbac_filter('__health_diagnostics_dismissed', 'health_diagnostics_dismissed', 'hostkey');

GRANT ALL PRIVILEGES ON SEQUENCE events_id_seq TO public;
GRANT ALL PRIVILEGES ON TABLE events TO public;
